import contextlib


def get_context(seed):
    context = {}
    context["seed"] = seed
    return context


@contextlib.contextmanager
def update_context(ctx, **kwargs):
    old_ctx = {key: ctx.get(key) for key in kwargs}
    ctx.update(kwargs)
    yield ctx
    ctx.update(old_ctx)



import pathlib
from copy import deepcopy

import msgpack
import yaml


class ModuleNotFoundError(Exception):
    pass


class FileStorage:
    def __init__(self, path):
        self.path = pathlib.Path(path)

    def list_modules(self):
        return sorted(fn.stem for fn in self.path.glob("*.yaml"))

    def resolve_module(self, name):
        try:
            with open(self.path / f"{name}.yaml") as fi:
                return yaml.safe_load(fi)
        except FileNotFoundError:
            raise ModuleNotFoundError(name)


class CompilingFileStorage(FileStorage):
    def clean(self):
        for fn in self.path.glob("*.mp"):
            fn.remove()

    def recompile_modules(self):
        self.clean()
        for fn in self.path.glob("*.yaml"):
            self.resolve_module(fn.stem)

    def resolve_module(self, name):
        compiled_fn = self.path / f"{name}.mp"
        try:
            with open(compiled_fn, "rb") as fi:
                return msgpack.load(fi, raw=False)
        except (FileNotFoundError, TypeError, ValueError, msgpack.ExtraData):
            mod = super().resolve_module(name)
            with open(compiled_fn, "wb") as fo:
                msgpack.dump(mod, fo)
            return mod


class CachedFileStorage(FileStorage):
    _modules = {}

    def resolve_module(self, name):
        try:
            return deepcopy(self._modules[name])
        except KeyError:
            mod = super().resolve_module(name)
            self._modules[name] = mod
            return self.resolve_module(name)


class CompilingCachedFileStorage(CompilingFileStorage):
    _modules = {}

    def resolve_module(self, name):
        try:
            return msgpack.loads(self._modules[name], raw=False)
        except KeyError:
            mod = super().resolve_module(name)
            self._modules[name] = msgpack.dumps(mod)
            return self.resolve_module(name)



import textwrap

from funcy import is_list, is_mapping, isa

from . import db, texts

is_bool = isa(bool)


def parse_grammar_file(storage, grammar_path, context, included=None):
    if included is None:
        included = {grammar_path}
    else:
        if grammar_path in included:
            return context
        else:
            included.add(grammar_path)

    doc = storage.resolve_module(grammar_path)
    context = parse_includes(
        storage, grammar_path, doc.pop("include", ()), context, included
    )

    render_strategy = parse_render_strategy(
        doc.pop("render", "ftemplate"), grammar_path
    )

    return parse_data(doc, grammar_path, context, render_strategy=render_strategy)


def parse_render_strategy(render_mode, grammar_path):
    if render_mode == "ftemplate":
        return texts.render_ftemplate
    elif render_mode == "jinja2" or render_mode == "jinja":
        return texts.render_jinja2
    else:
        raise ValueError(
            f"Unrecognized render strategy `{render_mode}` in {grammar_path}"
        )


def parse_includes(storage, grammar_path, includes, context, included):
    for include in includes:
        context = parse_grammar_file(storage, include, context, included)
    return context


def parse_data(data, grammar_path, context, render_strategy=texts.render_ftemplate):
    for key, value in data.items():
        context[key] = parse_value(
            value, f"{grammar_path}:{key}", context, render_strategy=render_strategy
        )

    return context


def get_list_setting(value, grammar_path):
    mode = "reuse"
    if value and is_mapping(value[0]):
        if len(value[0]) == 1 and "mode" in value[0]:
            mode = value.pop(0)["mode"]
    return mode, value


def parse_value(value, grammar_path, context, render_strategy=texts.render_ftemplate):
    if is_list(value):
        mode, value = get_list_setting(value, grammar_path)
        if mode == "reuse":
            return db.Database(
                db.choose(
                    [
                        parse_value(
                            v, grammar_path, context, render_strategy=render_strategy
                        )
                        for v in value
                    ]
                ),
                grammar_path,
                context,
            )
        elif mode == "pick":
            return db.Database(
                db.pick(
                    [
                        parse_value(
                            v, grammar_path, context, render_strategy=render_strategy
                        )
                        for v in value
                    ]
                ),
                grammar_path,
                context,
            )
        elif mode == "markov":
            return db.Database(
                db.markovify([texts.RenderedStr(v) for v in value]),
                grammar_path,
                context,
            )
        elif mode == "ratchet":
            return db.Database(
                db.ratchet([texts.RenderedStr(v) for v in value]), grammar_path, context
            )
        elif mode == "list":
            return db.Datalist(
                grammar_path, context, [texts.RenderedStr(v) for v in value]
            )
    elif is_mapping(value):
        return db.Databag(
            grammar_path,
            context,
            {
                k: parse_value(
                    v, grammar_path, context, render_strategy=render_strategy
                )
                for k, v in value.items()
            },
        )
    elif is_bool(value):
        return value
    else:
        return texts.RenderableText(
            fix_text(value), grammar_path, context, render_strategy=render_strategy
        )


def fix_text(text):
    return textwrap.dedent(text).strip()



import logging

from . import contexts, grammars, seeds


def render_story(storage, name, start="Begin", **kwargs):
    logging.debug(f"Rendering {name} w/ {storage}")
    ctx = contexts.get_context(seeds.make_seed())
    ctx.update(kwargs)
    ctx = grammars.parse_grammar_file(storage, name, ctx)
    return str(ctx[start])



import os
import sys

from setuptools import setup

sys.path.insert(0, os.path.dirname(__file__))

import versioneer  # noqa: E402

setup(version=versioneer.get_version(), cmdclass=versioneer.get_cmdclass())



import collections
import random

from . import seeds


class MarkovChainDict(collections.abc.Mapping):
    """A Markov Chain dictionary, for generating random strings.

    Derived from Peter Corbett's CGI random name generator:
    http://www.pick.ucam.org/~ptc24/mchain.html
    """

    def __init__(self):
        self._dict = collections.defaultdict(list)

    def __getitem__(self, key):
        return self._dict[key]

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def add_key(self, prefix, suffix):
        self._dict[prefix].append(suffix)

    def get_suffix(self, prefix, rng=random):
        return rng.choice(self[prefix])


class NameGenerator(collections.abc.Iterator):
    """Uses a Markov Chain to generate random names.

    Derived from Peter Corbett's CGI random name generator, with input
    from the ElderLore object-oriented variation.

    http://www.pick.ucam.org/~ptc24/mchain.html
    """

    def __init__(self, source_names, chainlen=2, seed=None):
        if 1 > chainlen > 10:
            raise ValueError("Chain length must be between 1 and 10, inclusive")
        self.chainlen = chainlen
        self.markov = MarkovChainDict()
        self.read_data(source_names, chainlen)

    def __next__(self):
        return self.get_random_name()

    def read_data(self, names, destroy=False):
        if destroy:
            del self.markov
            self.markov = MarkovChainDict()

        oldnames = []
        chainlen = self.chainlen

        for name in names:
            oldnames.append(name)
            spacer = "".join((" " * chainlen, name))
            name_len = len(name)
            for num in range(name_len):
                self.markov.add_key(
                    spacer[num : num + chainlen], spacer[num + chainlen]
                )
            self.markov.add_key(spacer[name_len : name_len + chainlen], "\n")

    def get_random_name(self, start="", max_length=10, seed=None):
        """Return a random name."""
        rng = seeds.get_rng(seed)
        prefix = start[-self.chainlen :] or " " * self.chainlen
        name = start
        suffix = ""
        while 1:
            suffix = self.markov.get_suffix(prefix, rng=rng)
            if suffix == "-":
                continue
            elif suffix == "\n" or len(name) == max_length:
                break
            else:
                name = "".join((name, suffix))
                prefix = "".join((prefix[1:], suffix))
        return name
